/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */

package org.netbeans.modules.cnd.dwarfdump.dwarf;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

/**
 *
 * @author ak119685
 */
public class DwarfAbbriviationTable {
    private ArrayList<DwarfAbbriviationTableEntry> entries = null;
    private int numOfEntries = 0;
    @SuppressWarnings("unused")
	private long offset;
    private boolean consecutive;
    
    private static class EntryComparator implements Comparator<DwarfAbbriviationTableEntry> {
        public int compare(DwarfAbbriviationTableEntry x, DwarfAbbriviationTableEntry y) {
            return (int)x.getTableIndex() - (int)y.getTableIndex();
        }
    }

    public DwarfAbbriviationTable(long offset) {
        this.offset = offset;
    }
    
    public void setEntries(ArrayList<DwarfAbbriviationTableEntry> entries) {
        this.entries = entries;
        this.numOfEntries = entries.size();
        Collections.sort(entries, new EntryComparator());
        if (numOfEntries > 0) {
        	consecutive = entries.get(0).getTableIndex() == 1 &&
        		entries.get(numOfEntries - 1).getTableIndex() == numOfEntries;
        } else {
            consecutive = true;
        }
    }
    
    public DwarfAbbriviationTableEntry getEntry(long idx) {
    	if (consecutive) {
    		if (idx > 0 && idx <= numOfEntries) {
    			return entries.get((int)idx - 1);
    		}
    	} else {
    		// TODO could use binary search here.
    		for (DwarfAbbriviationTableEntry entry : entries) {
    			if (idx == entry.getTableIndex()) {
    				return entry;
    			} else if (idx < entry.getTableIndex()) {
    				break;
    			}
    		}
    	}
    	// Not found
    	return null;
    }
    
    public void dump(PrintStream out) {
        out.println("  Number TAG"); // NOI18N
        
        for (DwarfAbbriviationTableEntry entry : entries) {
            entry.dump(out);
        }
    }
}
